@@ -1,4 +1,4 @@
-from cStringIO import StringIO
+from io import StringIO
 import sys
 import unittest
 
@@ -11,8 +11,8 @@ from Bio import triefind
 
 IS_PYPY = hasattr(sys, 'pypy_version_info')
 
-class QualityFilterTestCase(unittest.TestCase):
 
+class QualityFilterTestCase(unittest.TestCase):
     def setUp(self):
         self.instance = quality_filter.QualityScoreFilter(25.0)
         self.sequence = SeqRecord(Seq('ACGT'))
@@ -20,16 +20,17 @@ class QualityFilterTestCase(unittest.TestCase):
     def test_nowindow_fail(self):
         self.sequence.letter_annotations['phred_quality'] = [25, 25, 24, 25]
         instance = quality_filter.QualityScoreFilter()
-        self.assertRaises(quality_filter.FailedFilter, instance.filter_record, self.sequence)
+        self.assertRaises(quality_filter.FailedFilter, instance.filter_record,
+                          self.sequence)
 
     def test_nowindow_pass(self):
         self.sequence.letter_annotations['phred_quality'] = [25, 25, 25, 25]
         instance = quality_filter.QualityScoreFilter()
         result = instance.filter_record(self.sequence)
-        self.assertEqual(self.sequence, result)
+        self.assertEqual(self.sequence.seq, result.seq)
+
 
 class WindowQualityFilterTestCase(unittest.TestCase):
-
     def setUp(self):
         self.instance = quality_filter.WindowQualityScoreFilter(2, 25)
         self.sequence = SeqRecord(Seq('ACGT'))
@@ -41,7 +42,8 @@ class WindowQualityFilterTestCase(unittest.TestCase):
 
     def test_window_truncate_noseq(self):
         self.sequence.letter_annotations['phred_quality'] = [25, 24, 25, 25]
-        self.assertRaises(quality_filter.FailedFilter, self.instance.filter_record, self.sequence)
+        self.assertRaises(quality_filter.FailedFilter,
+                          self.instance.filter_record, self.sequence)
 
     def test_window_truncate_mid(self):
         self.sequence.letter_annotations['phred_quality'] = [25, 25, 23, 25]
@@ -49,17 +51,20 @@ class WindowQualityFilterTestCase(unittest.TestCase):
         self.assertEqual(2, len(result))
         self.assertEqual('AC', str(result.seq))
 
+
 class AmbiguousBaseFilterTestCase(unittest.TestCase):
     """
     Tests for ambiguous_base_filter
     """
+
     def setUp(self):
-        self.records = [SeqRecord(Seq('ACGT')),
-                SeqRecord(Seq('NNNN')),
-                SeqRecord(Seq('NACT')),
-                SeqRecord(Seq('ACGTN')),
-                SeqRecord(Seq('GGNTTACT')),
-                ]
+        self.records = [
+            SeqRecord(Seq('ACGT')),
+            SeqRecord(Seq('NNNN')),
+            SeqRecord(Seq('NACT')),
+            SeqRecord(Seq('ACGTN')),
+            SeqRecord(Seq('GGNTTACT')),
+        ]
 
     def test_drop(self):
         """
@@ -70,7 +75,7 @@ class AmbiguousBaseFilterTestCase(unittest.TestCase):
         self.assertEqual(1, len(actual))
         self.assertEqual(1, instance.passed)
         self.assertEqual(4, instance.failed)
-        self.assertEqual(self.records[0], actual[0])
+        self.assertEqual(self.records[0].seq, actual[0].seq)
 
     def test_truncate(self):
         instance = quality_filter.AmbiguousBaseFilter('truncate')
@@ -79,22 +84,22 @@ class AmbiguousBaseFilterTestCase(unittest.TestCase):
         self.assertEqual(0, instance.failed)
         self.assertEqual(5, instance.passed)
         self.assertEqual(['ACGT', '', '', 'ACGT', 'GG'],
-                [str(s.seq) for s in actual])
+                         [str(s.seq) for s in actual])
 
     def test_invalid_action(self):
         self.assertRaises(ValueError, quality_filter.AmbiguousBaseFilter,
-                'other')
+                          'other')
+
 
 class MaxAmbiguousFilterTestCase(unittest.TestCase):
-    """
-    """
     def setUp(self):
-        self.records = [SeqRecord(Seq('ACGT')),
-                SeqRecord(Seq('NNNN')),
-                SeqRecord(Seq('NACT')),
-                SeqRecord(Seq('ACNTN')),
-                SeqRecord(Seq('GGNTTNACT')),
-                ]
+        self.records = [
+            SeqRecord(Seq('ACGT')),
+            SeqRecord(Seq('NNNN')),
+            SeqRecord(Seq('NACT')),
+            SeqRecord(Seq('ACNTN')),
+            SeqRecord(Seq('GGNTTNACT')),
+        ]
 
     def test_none(self):
         instance = quality_filter.MaxAmbiguousFilter(0)
@@ -113,11 +118,40 @@ class MaxAmbiguousFilterTestCase(unittest.TestCase):
         self.assertEqual([self.records[i] for i in (0, 2)], filtered)
 
 
-class MinLengthFilterTestCase(unittest.TestCase):
-
+class PctAmbiguousFilterTestCase(unittest.TestCase):
     def setUp(self):
-        self.sequences = [SeqRecord(Seq('ACGT')),
-                          SeqRecord(Seq('ACTTT')), ]
+        self.records = [
+            SeqRecord(Seq('ACGT')),
+            SeqRecord(Seq('NNNN')),
+            SeqRecord(Seq('NACT')),
+            SeqRecord(Seq('ACNTN')),
+            SeqRecord(Seq('GGNTTNACT')),
+        ]
+
+    def test_none(self):
+        instance = quality_filter.PctAmbiguousFilter(0)
+        filtered = list(instance.filter_records(self.records))
+        self.assertEqual(len(filtered), 1)
+        self.assertEqual(str(self.records[0].seq), str(filtered[0].seq))
+
+    def test_10(self):
+        instance = quality_filter.PctAmbiguousFilter(100)
+        filtered = list(instance.filter_records(self.records))
+        self.assertEqual(filtered, self.records)
+
+    def test_1(self):
+        instance = quality_filter.PctAmbiguousFilter(0.23)
+        filtered = list(instance.filter_records(self.records))
+        print(filtered)
+        self.assertEqual([self.records[i] for i in (0, 4)], filtered)
+
+
+class MinLengthFilterTestCase(unittest.TestCase):
+    def setUp(self):
+        self.sequences = [
+            SeqRecord(Seq('ACGT')),
+            SeqRecord(Seq('ACTTT')),
+        ]
 
     def test_none_pass(self):
         instance = quality_filter.MinLengthFilter(6)
@@ -134,10 +168,13 @@ class MinLengthFilterTestCase(unittest.TestCase):
         actual = list(instance.filter_records(self.sequences))
         self.assertEqual(self.sequences[1:], actual)
 
+
 class MaxLengthFilterTestCase(unittest.TestCase):
     def setUp(self):
-        self.sequences = [SeqRecord(Seq('ACGT')),
-                          SeqRecord(Seq('ACTTT')), ]
+        self.sequences = [
+            SeqRecord(Seq('ACGT')),
+            SeqRecord(Seq('ACTTT')),
+        ]
 
     def test_none_truncated(self):
         instance = quality_filter.MaxLengthFilter(6)
@@ -153,33 +190,39 @@ class MaxLengthFilterTestCase(unittest.TestCase):
         instance = quality_filter.MaxLengthFilter(3)
         actual = list(instance.filter_records(self.sequences))
         self.assertEqual(['ACG', 'ACT'], [str(s.seq) for s in actual])
-        self.assertEqual([i.id for i in self.sequences], [i.id for i in actual])
+        self.assertEqual([i.id for i in self.sequences],
+                         [i.id for i in actual])
+
 
 @unittest.skipIf(IS_PYPY, "Bio.trie not available on pypy.")
 class PrimerBarcodeFilterTestCase(unittest.TestCase):
     def setUp(self):
-        self.sequences = [SeqRecord(Seq('ACCGTTACGAT'), 'seq1'),
-                          SeqRecord(Seq('ACTGTTACGCT'), 'seq2'),
-                          SeqRecord(Seq('AACTGTTA'), 'seq3'), # Homopolymer in bc
-                          SeqRecord(Seq('ACCGTA'), 'seq4'),   # Error in primer
-                          ]
+        self.sequences = [
+            SeqRecord(Seq('ACCGTTACGAT'), 'seq1'),
+            SeqRecord(Seq('ACTGTTACGCT'), 'seq2'),
+            SeqRecord(Seq('AACTGTTA'), 'seq3'),  # Homopolymer in bc
+            SeqRecord(Seq('ACCGTA'), 'seq4'),  # Error in primer
+        ]
 
         barcode_str = """Sample1,ACC\nSample2,ACT\n"""
         self.primer = 'GTTA'
-        self.trie = quality_filter.parse_barcode_file(StringIO(barcode_str), primer=self.primer)
+        self.trie = quality_filter.parse_barcode_file(
+            StringIO(barcode_str), primer=self.primer)
         self.outfile = StringIO()
-        self.instance = quality_filter.PrimerBarcodeFilter(self.trie, self.outfile)
+        self.instance = quality_filter.PrimerBarcodeFilter(
+            self.trie, self.outfile)
 
     def test_filter_trim(self):
         actual = list(self.instance.filter_records(self.sequences))
         self.assertEqual(2, len(actual))
         self.assertEqual(['CGAT', 'CGCT'], [str(s.seq) for s in actual])
 
-class RecordEventListenerTestCase(unittest.TestCase):
 
+class RecordEventListenerTestCase(unittest.TestCase):
     def test_send(self):
         events = []
         record = object()
+
         def e_handler(record, n=1):
             events.append(n)
 
@@ -196,9 +239,9 @@ class RecordEventListenerTestCase(unittest.TestCase):
         rle('other', record, n=5)
         self.assertEqual(events, [1, 5])
 
+
 @unittest.skipIf(IS_PYPY, "Bio.trie not available on pypy.")
 class BarcodePrimerTrieTestCase(unittest.TestCase):
-
     def setUp(self):
         self.barcode_str = """p1d1bc205,TACTAGCG,CATTGCCTATG
 p1d1bc206,TACTCGTC,CATTGCCTATG
@@ -213,14 +256,15 @@ p1d1bc213,TACGAGAC,CAYGGCTA"""
 
     def test_primer_provided(self):
         res = quality_filter.parse_barcode_file(self.fp, primer='CATTGCCTATG')
-        self.assertEqual(9, len(res.keys()))
+        self.assertEqual(9, len(list(res.keys())))
         self.assertEqual('p1d1bc210', res['TACAGTCGCATTGCCTATG'])
         self.assertEqual(None, triefind.match('TACAGTCGCATTGCCTAT', res))
-        self.assertEqual('TACAGTCGCATTGCCTATG', triefind.match('TACAGTCGCATTGCCTATGCTACCTA', res))
+        self.assertEqual('TACAGTCGCATTGCCTATG',
+                         triefind.match('TACAGTCGCATTGCCTATGCTACCTA', res))
 
     def test_primer_in_file(self):
         res = quality_filter.parse_barcode_file(self.fp, primer=None)
-        self.assertEqual(13, len(res.keys()))
+        self.assertEqual(13, len(list(res.keys())))
 
         # Test ambiguities
         self.assertEqual('p1d1bc212', res['TACGTCTCCATGGCTA'])
@@ -228,12 +272,12 @@ p1d1bc213,TACGAGAC,CAYGGCTA"""
         self.assertIsNone(res.get('TACGTCTCCAAGGCTA'))
         self.assertIsNone(res.get('TACGTCTCCAGGGCTA'))
 
+
 class AllUnambiguousTestCase(unittest.TestCase):
     def test_one_nt(self):
         self.assertEqual(set('ACGT'), set(quality_filter.all_unambiguous('N')))
 
     def test_four_nt(self):
-        self.assertEqual(set(['ACCG', 'ACCA']),
-                         set(quality_filter.all_unambiguous('ACCR')))
+        self.assertEqual(
+            set(['ACCG', 'ACCA']), set(quality_filter.all_unambiguous('ACCR')))
         self.assertEqual(4**4, len(quality_filter.all_unambiguous('NNNN')))
-
