@@ -2,7 +2,7 @@
 Tests for seqmagick.transform
 """
 
-from cStringIO import StringIO
+from io import StringIO
 import functools
 import logging
 import unittest
@@ -19,20 +19,23 @@ def _alignment_record(sequence):
     return SeqRecord(Seq(sequence,
         alphabet=Alphabet.Gapped(Alphabet.generic_dna)))
 
-def seqrecord(sequence_id, sequence_text, alphabet=Alphabet.generic_dna):
+def seqrecord(sequence_id, sequence_text, alphabet=Alphabet.generic_dna,
+              description=None):
     """
     Quick shortcut to make a SeqRecord
     """
-    return SeqRecord(Seq(sequence_text, alphabet), id=sequence_id)
+    record = SeqRecord(Seq(sequence_text, alphabet),
+                       id=sequence_id)
+    if description:
+        record.description = description
+    return record
 
 class PatternReplaceTestCase(unittest.TestCase):
 
     def create_sequences(self):
-        return [
-    seqrecord('test_sequence_1', 'ACTGT'),
-    seqrecord('test_REPLACE_2', 'ACTGT'),
-    seqrecord('other_sequence', 'ATGAG'),
-    ]
+        return [seqrecord('test_sequence_1', 'ACTGT'),
+                seqrecord('test_REPLACE_2', 'ACTGT'),
+                seqrecord('other_sequence', 'ATGAG'), ]
 
     def setUp(self):
         super(PatternReplaceTestCase, self).setUp()
@@ -41,6 +44,81 @@ class PatternReplaceTestCase(unittest.TestCase):
     def tearDown(self):
         super(PatternReplaceTestCase, self).tearDown()
 
+    # from http://stackoverflow.com/questions/13923072/shortening-fasta-header-perl
+    def test_pattern_replace_anchored_transform_id(self):
+        sequences = [seqrecord('gi|351517969|ref|NW_003613580.1|', 'CAGTC',
+                               description='gi|351517969|ref|NW_003613580.1| Cricetulus griseus unplaced genomic scaffold'),
+                    seqrecord('gi|351517969|ref|NW_003613580.1|', 'CAGTC',
+                               description='gi|351517969|ref|NW_003613580.1|'),
+                    seqrecord('gi|351517969|ref|NW_003613580.1|', 'CAGTC')]
+
+        # capture the identifier after three groups of pipe-separated characters
+        transformed = list(transform.name_replace(sequences, r'^(?:[^|]+\|){3}([^|]+)\|', r'\1'))
+
+        self.assertEqual(str(sequences[0].seq), str(transformed[0].seq))
+        self.assertEqual('NW_003613580.1', transformed[0].id)
+        self.assertEqual('NW_003613580.1 Cricetulus griseus unplaced genomic scaffold', transformed[0].description)
+
+        self.assertEqual(str(sequences[1].seq), str(transformed[1].seq))
+        self.assertEqual('NW_003613580.1', transformed[1].id)
+        self.assertEqual('NW_003613580.1', transformed[1].description)
+
+        self.assertEqual(str(sequences[2].seq), str(transformed[2].seq))
+        self.assertEqual('NW_003613580.1', transformed[2].id)
+        self.assertEqual('<unknown description>', transformed[2].description)
+
+    # from http://stackoverflow.com/questions/15155728/modifying-fasta-headers-with-unix-command-line-tools
+    def test_pattern_replace_anchored_id_from_description(self):
+        sequences = [seqrecord('hg19_ct_UserTrack_3545_691', 'GATGG',
+                               description='hg19_ct_UserTrack_3545_691 range=chr1:8121498-8121502 5\'pad=0 3\'pad=0 strand=+ repeatMasking=none')]
+
+        transformed = next(transform.name_replace(sequences, r'^\S+ range=(\S+)', r'\1'))
+
+        self.assertEqual(str(sequences[0].seq), str(transformed.seq))
+        self.assertEqual('chr1:8121498-8121502', transformed.id)
+        self.assertEqual('chr1:8121498-8121502 5\'pad=0 3\'pad=0 strand=+ repeatMasking=none', transformed.description)
+
+    # from http://stackoverflow.com/questions/23280240/how-to-rename-fasta-file-headers-using-sed
+    def test_pattern_replace_anchored_add_to_description(self):
+        sequences = [seqrecord('Bra000001', 'CTTAT', description='Bra000001')]
+
+        transformed = next(transform.name_replace(sequences, r'^(Bra\d+)$', r'\1 Brassica rapa'))
+
+        self.assertEqual(str(sequences[0].seq), str(transformed.seq))
+        self.assertEqual('Bra000001', transformed.id)
+        self.assertEqual('Bra000001 Brassica rapa', transformed.description)
+
+    def test_pattern_replace_anchored_remove_from_description(self):
+        sequences = [seqrecord('Bra000001', 'CTTAT', description='Bra000001 Brassica rapa')]
+
+        transformed = next(transform.name_replace(sequences, r' .*$', ''))
+
+        self.assertEqual(str(sequences[0].seq), str(transformed.seq))
+        self.assertEqual('Bra000001', transformed.id)
+        self.assertEqual('Bra000001', transformed.description)
+
+    def test_pattern_replace_anchored_nomatch(self):
+        sequences = [seqrecord('hello', 'A', description='hello friend')]
+        transformed = next(transform.name_replace(sequences, r'^hello$', 'bye'))
+
+        self.assertEqual(str(sequences[0].seq), str(transformed.seq))
+        self.assertEqual('hello', transformed.id)
+        self.assertEqual('hello friend', transformed.description)
+
+    def test_pattern_replace_anchored_match(self):
+        sequences = [seqrecord('hello', 'A', description='hello friend'),
+                     seqrecord('hello', 'A')]
+
+        transformed = list(transform.name_replace(sequences, r'^hello\b', 'bye'))
+
+        self.assertEqual(str(sequences[0].seq), str(transformed[0].seq))
+        self.assertEqual('bye', transformed[0].id)
+        self.assertEqual('bye friend', transformed[0].description)
+
+        self.assertEqual(str(sequences[1].seq), str(transformed[1].seq))
+        self.assertEqual('bye', transformed[1].id)
+        self.assertEqual('<unknown description>', transformed[1].description)
+
     def test_pattern_replace_none(self):
         result = transform.name_replace(self.sequences, 'ZZZ', 'MATCH')
         result = list(result)
@@ -120,29 +198,34 @@ class SeqPatternTestCase(unittest.TestCase):
         super(SeqPatternTestCase, self).setUp()
 
         self.sequences = [
-            seqrecord('sequence_1', 'AC-G--'),
-            seqrecord('sequence_2', '-C-GT-'),
-            seqrecord('sequence_3', '-T-AG-'),
+            seqrecord('s1', 'AC-G--'),
+            seqrecord('s2', '-C-GT-'),
+            seqrecord('s3', '-T-AG-'),
         ]
 
-        self.tests = [('^$', []), ('.*', self.sequences),
-                ('^AC', [self.sequences[0]])]
+        self.tests = [('^$', set()),
+                      ('.*', {'s1', 's2', 's3'}),
+                      ('^AC', {'s1'}),
+                      ('^ac', set()),
+                      ('(?i)^ac', {'s1'})]
 
     def test_include(self):
         result = transform.seq_include(self.sequences, '^$')
 
         for regex, expected in self.tests:
-            result = list(transform.seq_include(self.sequences, regex))
+            result = {seq.id for seq in transform.seq_include(self.sequences, regex)}
             self.assertEqual(expected, result)
 
     def test_exclude(self):
         result = transform.seq_include(self.sequences, '^$')
 
         for regex, expected_include in self.tests:
-            expected = [i for i in self.sequences if i not in expected_include]
-            result = list(transform.seq_exclude(self.sequences, regex))
+            result = {seq.id for seq in transform.seq_exclude(self.sequences, regex)}
+            expected = {seq.id for seq in self.sequences
+                        if seq.id not in expected_include}
             self.assertEqual(expected, result)
 
+
 class HeadTestCase(unittest.TestCase):
     """
     Test for transform.head
@@ -150,10 +233,10 @@ class HeadTestCase(unittest.TestCase):
 
     def setUp(self):
         self.sequences = [seqrecord('sequence{0}'.format(i), 'A'*(i+1))
-                          for i in xrange(100)]
+                          for i in range(100)]
 
     def test_zero(self):
-        result = list(transform.head(self.sequences, 0))
+        result = list(transform.head(self.sequences, '0'))
         self.assertEqual([], result)
 
     def test_more_seqs_than_available(self):
@@ -161,17 +244,45 @@ class HeadTestCase(unittest.TestCase):
         Specifying more sequences than are in input records should return
         them all
         """
-        result = list(transform.head(self.sequences, 10000))
-        self.assertEqual(self.sequences, result)
+        result = list(transform.head(self.sequences, '10000'))
+        self.assertEqual([s.id for s in self.sequences],
+                         [r.id for r in result])
+        self.assertEqual([str(s.seq) for s in self.sequences],
+                         [str(r.seq) for r in result])
 
     def test_values(self):
         """
         Try specifying some values.
         """
-        for h in xrange(len(self.sequences) + 1):
-            result = list(transform.head(self.sequences, h))
+        for h in range(len(self.sequences) + 1):
+            result = list(transform.head(self.sequences, str(h)))
             self.assertEqual(h, len(result))
-            self.assertEqual(self.sequences[:h], result)
+            self.assertEqual([s.id for s in self.sequences[:h]],
+                             [r.id for r in result])
+            self.assertEqual([str(s.seq) for s in self.sequences[:h]],
+                             [str(r.seq) for r in result])
+
+    def test_minus_zero(self):
+        """
+        Test that -0 returns all sequences
+        """
+        result = list(transform.head(self.sequences, '-0'))
+        self.assertEqual([s.id for s in self.sequences],
+                         [r.id for r in result])
+        self.assertEqual([str(s.seq) for s in self.sequences],
+                         [str(r.seq) for r in result])
+
+    def test_minus_values(self):
+        """
+        Try specifying some minus values.
+        """
+        for h in range(1, len(self.sequences) + 1):
+            result = list(transform.head(self.sequences, str(-h)))
+            self.assertEqual(h, len(self.sequences) - len(result))
+            self.assertEqual([s.id for s in self.sequences[:-h]],
+                             [r.id for r in result])
+            self.assertEqual([str(s.seq) for s in self.sequences[:-h]],
+                             [str(r.seq) for r in result])
 
 class TailTestCase(unittest.TestCase):
     def setUp(self):
@@ -182,7 +293,7 @@ class TailTestCase(unittest.TestCase):
         ]
 
     def _do_test(self, size):
-        actual = list(transform.tail(self.records, size))
+        actual = list(transform.tail(self.records, str(size)))
         expected = self.records[-size:]
         self.assertEqual([e.id for e in expected], [a.id for a in actual])
         self.assertEqual([str(e.seq) for e in expected], [str(a.seq) for a in actual])
@@ -196,6 +307,28 @@ class TailTestCase(unittest.TestCase):
     def test_tail_3(self):
         self._do_test(3)
 
+    def test_plus_zero(self):
+        """
+        Test that +0 returns all sequences
+        """
+        result = list(transform.tail(self.records, '+0'))
+        self.assertEqual([s.id for s in self.records],
+                         [r.id for r in result])
+        self.assertEqual([str(s.seq) for s in self.records],
+                         [str(r.seq) for r in result])
+
+    def test_plus_values(self):
+        """
+        Try specifying some plus values.
+        """
+        for h in range(1, len(self.records) + 1):
+            result = list(transform.tail(self.records, '+{}'.format(h)))
+            self.assertEqual(len(self.records) + 1 - h, len(result))
+            self.assertEqual([s.id for s in self.records[h-1:]],
+                             [r.id for r in result])
+            self.assertEqual([str(s.seq) for s in self.records[h-1:]],
+                             [str(r.seq) for r in result])
+
 class IsolateRegionTestCase(unittest.TestCase):
 
     def setUp(self):
@@ -226,9 +359,9 @@ class IsolateRegionTestCase(unittest.TestCase):
 
     def test_invalid(self):
         self.assertRaises(ValueError, transform.isolate_region(
-                self.sequences, 5, 5).next)
+                self.sequences, 5, 5).__next__)
         self.assertRaises(ValueError, transform.isolate_region(
-                self.sequences, 10, 5).next)
+                self.sequences, 10, 5).__next__)
 
 class MinUngapLengthTestCase(unittest.TestCase):
 
@@ -248,7 +381,10 @@ class MinUngapLengthTestCase(unittest.TestCase):
 
     def test_partial(self):
         result = transform.min_ungap_length_discard(self.sequences, 4)
-        self.assertEqual([self.sequences[1], self.sequences[3]], list(result))
+        self.assertEqual(
+            [self.sequences[1].seq, self.sequences[3].seq],
+            [seq.seq for seq in result])
+
 
 class IncludeExcludeMixIn(object):
 
@@ -259,11 +395,13 @@ sequenceid4
 """
         self.handle = StringIO(ids)
 
-        self.sequences = [SeqRecord(Seq("AAA"), id="sequenceid1"),
-                SeqRecord(Seq("BBB"), id="sequenceid2"),
-                SeqRecord(Seq("CCC"), id="sequenceid3"),
-                SeqRecord(Seq("DDD"), id="sequenceid4", description='sequence id 4'),
-                SeqRecord(Seq("EEE"), id="test", description='test sequence'), ]
+        self.sequences = [seqrecord("sequenceid1", "AAA"),
+                          seqrecord("sequenceid2", "BBB"),
+                          seqrecord("sequenceid3", "CCC"),
+                          seqrecord("sequenceid4", "DDD",
+                                    description='sequence id 4'),
+                          seqrecord("test", "EEE",
+                                    description='test sequence'), ]
 
 
 class IncludeFromFileTestCase(IncludeExcludeMixIn, unittest.TestCase):
@@ -343,11 +481,11 @@ class CodonWarningTableTestCase(unittest.TestCase):
     def setUp(self):
         self.warnings = []
         self.warning_dict = transform.CodonWarningTable({'UUU': 'F'})
-        self.old_warn = transform.logging.warn
-        transform.logging.warn = self.warn
+        self.old_warn = transform.logging.warning
+        transform.logging.warning = self.warn
 
     def tearDown(self):
-        transform.logging.warn = self.old_warn
+        transform.logging.warning = self.old_warn
 
     def test_nowarn(self):
         actual = self.warning_dict['UUU']
