@@ -3,8 +3,8 @@ Functions to transform / filter sequences
 """
 import collections
 import contextlib
-import copy
-import cPickle as pickle
+import pickle as pickle
+import gzip
 import itertools
 import logging
 import re
@@ -18,12 +18,15 @@ from Bio.Data import CodonTable
 from Bio.Seq import Seq
 from Bio.SeqRecord import SeqRecord
 from Bio.SeqUtils.CheckSum import seguid
+from functools import reduce
 
 # Characters to be treated as gaps
 GAP_CHARS = "-."
+GAP_TABLE = {ord(c): None for c in GAP_CHARS}
+
+# Size of temporary file buffer: default to 256MB
+DEFAULT_BUFFER_SIZE = 268435456  # 256 * 2**20
 
-# Size of temporary file buffer: default to 20MB
-DEFAULT_BUFFER_SIZE = 20971520 # 20*2**20
 
 @contextlib.contextmanager
 def _record_buffer(records, buffer_size=DEFAULT_BUFFER_SIZE):
@@ -55,8 +58,9 @@ def dashes_cleanup(records, prune_chars='.:?~'):
     Take an alignment and convert any undesirable characters such as ? or ~ to
     -.
     """
-    logging.info("Applying _dashes_cleanup: converting any . or : to -.")
-    translation_table = string.maketrans(prune_chars, '-' * len(prune_chars))
+    logging.info(
+        "Applying _dashes_cleanup: converting any of '{}' to '-'.".format(prune_chars))
+    translation_table = {ord(c): '-' for c in prune_chars}
     for record in records:
         record.seq = Seq(str(record.seq).translate(translation_table),
                          record.seq.alphabet)
@@ -81,7 +85,7 @@ def deduplicate_sequences(records, out_file):
 
     if out_file is not None:
         with out_file:
-            for sequences in checksum_sequences.itervalues():
+            for sequences in checksum_sequences.values():
                 out_file.write('%s\n' % (' '.join(sequences),))
 
 
@@ -185,7 +189,7 @@ def drop_columns(records, slices):
         # Generate a set of indices to remove
         drop = set(i for slice in slices
                    for i in range(*slice.indices(len(record))))
-        keep = [i not in drop for i in xrange(len(record))]
+        keep = [i not in drop for i in range(len(record))]
         record.seq = Seq(''.join(itertools.compress(record.seq, keep)), record.seq.alphabet)
         yield record
 
@@ -203,7 +207,7 @@ def multi_cut_sequences(records, slices):
             yield reduce(lambda x, y: x + y, pieces)
 
 def _update_slices(record, slices):
-    n = itertools.count().next
+    n = itertools.count().__next__
     # Generate a map from indexes in the specified sequence to those in the
     # alignment
     ungap_map = dict((n(), i) for i, base in enumerate(str(record.seq))
@@ -254,7 +258,7 @@ def multi_mask_sequences(records, slices):
     Replace characters sliced by slices with gap characters.
     """
     for record in records:
-        record_indices = range(len(record))
+        record_indices = list(range(len(record)))
         keep_indices = reduce(lambda i, s: i - frozenset(record_indices[s]),
                               slices, frozenset(record_indices))
         seq = ''.join(b if i in keep_indices else '-'
@@ -309,7 +313,7 @@ def _reverse_annotations(old_record, new_record):
     lists / tuples / strings.
     """
     # Copy the annotations over
-    for k, v in old_record.annotations.items():
+    for k, v in list(old_record.annotations.items()):
         # Trim if appropriate
         if isinstance(v, (tuple, list)) and len(v) == len(old_record):
             assert len(v) == len(old_record)
@@ -318,7 +322,7 @@ def _reverse_annotations(old_record, new_record):
 
     # Letter annotations must be lists / tuples / strings of the same
     # length as the sequence
-    for k, v in old_record.letter_annotations.items():
+    for k, v in list(old_record.letter_annotations.items()):
         assert len(v) == len(old_record)
         new_record.letter_annotations[k] = v[::-1]
 
@@ -355,20 +359,24 @@ def reverse_complement_sequences(records):
         yield rev_record
 
 
-def ungap_sequences(records, gap_chars=GAP_CHARS):
+def ungap_sequences(records, gap_chars=GAP_TABLE):
     """
     Remove gaps from sequences, given an alignment.
     """
-    logging.info('Applying _ungap_sequences generator: '
-                 'removing gaps from the alignment.')
+    logging.info('Applying _ungap_sequences generator: removing all gap characters')
     for record in records:
         yield ungap_all(record, gap_chars)
 
-def ungap_all(record, gap_chars=GAP_CHARS):
-    record = SeqRecord(Seq(str(record.seq).translate(None, gap_chars)),
-            id=record.id, description=record.description)
+
+def ungap_all(record, gap_chars=GAP_TABLE):
+
+    record = SeqRecord(
+        Seq(str(record.seq).translate(gap_chars)),
+        id=record.id, description=record.description
+    )
     return record
 
+
 def _update_id(record, new_id):
     """
     Update a record id to new_id, also modifying the ID in record.description
@@ -377,8 +385,7 @@ def _update_id(record, new_id):
     record.id = new_id
 
     # At least for FASTA, record ID starts the description
-    record.description = re.sub('^' + re.escape(old_id), new_id,
-            record.description)
+    record.description = re.sub('^' + re.escape(old_id), new_id, record.description)
     return record
 
 
@@ -417,7 +424,7 @@ def name_include(records, filter_regex):
     logging.info('Applying _name_include generator: '
                  'including only IDs matching ' + filter_regex +
                  ' in results.')
-    regex = re.compile(filter_regex, re.I)
+    regex = re.compile(filter_regex)
     for record in records:
         if regex.search(record.id) or regex.search(record.description):
             yield record
@@ -430,7 +437,7 @@ def name_exclude(records, filter_regex):
     """
     logging.info('Applying _name_exclude generator: '
                  'excluding IDs matching ' + filter_regex + ' in results.')
-    regex = re.compile(filter_regex, re.I)
+    regex = re.compile(filter_regex)
     for record in records:
         if not regex.search(record.id) and not regex.search(record.description):
             yield record
@@ -440,11 +447,22 @@ def name_replace(records, search_regex, replace_pattern):
     """
     Given a set of sequences, replace all occurrences of search_regex
     with replace_pattern. Ignore case.
+
+    If the ID and the first word of the description match, assume the
+    description is FASTA-like and apply the transform to the entire
+    description, then set the ID from the first word. If the ID and
+    the first word of the description do not match, apply the transform
+    to each individually.
     """
-    regex = re.compile(search_regex, re.I)
+    regex = re.compile(search_regex)
     for record in records:
-        record.id = regex.sub(replace_pattern, record.id)
-        record.description = regex.sub(replace_pattern, record.description)
+        maybe_id = record.description.split(None, 1)[0]
+        if maybe_id == record.id:
+            record.description = regex.sub(replace_pattern, record.description)
+            record.id = record.description.split(None, 1)[0]
+        else:
+            record.id = regex.sub(replace_pattern, record.id)
+            record.description = regex.sub(replace_pattern, record.description)
         yield record
 
 
@@ -452,7 +470,7 @@ def seq_include(records, filter_regex):
     """
     Filter any sequences who's seq does not match the filter. Ignore case.
     """
-    regex = re.compile(filter_regex, re.I)
+    regex = re.compile(filter_regex)
     for record in records:
         if regex.search(str(record.seq)):
             yield record
@@ -460,35 +478,77 @@ def seq_include(records, filter_regex):
 
 def seq_exclude(records, filter_regex):
     """
-    Filter any sequences who's seq matches the filter. Ignore case.
+    Filter any sequences whose seq matches the filter. Ignore case.
     """
-    regex = re.compile(filter_regex, re.I)
+    regex = re.compile(filter_regex)
     for record in records:
         if not regex.search(str(record.seq)):
             yield record
 
-def sample(records, sample):
-    return random.sample(list(records), sample)
+
+def sample(records, k, random_seed=None):
+    """Choose a length-``k`` subset of ``records``, retaining the input
+    order.  If k > len(records), all are returned. If an integer
+    ``random_seed`` is provided, sets ``random.seed()``
+
+    """
+
+    if random_seed is not None:
+        random.seed(random_seed)
+
+    result = []
+    for i, record in enumerate(records):
+        if len(result) < k:
+            result.append(record)
+        else:
+            r = random.randint(0, i)
+            if r < k:
+                result[r] = record
+    return result
+
 
 def head(records, head):
     """
     Limit results to the top N records.
+    With the leading `-', print all but the last N records.
     """
     logging.info('Applying _head generator: '
-                 'limiting results to top ' + str(head) + ' records.')
-    return itertools.islice(records, head)
+                 'limiting results to top ' + head + ' records.')
 
+    if head == '-0':
+        for record in records:
+            yield record
+    elif '-' in head:
+        with _record_buffer(records) as r:
+            record_count = sum(1 for record in r())
+            end_index = max(record_count + int(head), 0)
+            for record in itertools.islice(r(), end_index):
+                yield record
+    else:
+        for record in itertools.islice(records, int(head)):
+            yield record
 
 def tail(records, tail):
     """
     Limit results to the bottom N records.
+    Use +N to output records starting with the Nth.
     """
-    with _record_buffer(records) as r:
-        record_count = sum(1 for record in r())
-        start_index = record_count - tail
-        rec_iter = r()
-        for record in itertools.islice(rec_iter, start_index, None):
+    logging.info('Applying _tail generator: '
+                 'limiting results to top ' + tail + ' records.')
+
+    if tail == '+0':
+        for record in records:
             yield record
+    elif '+' in tail:
+        tail = int(tail) - 1
+        for record in itertools.islice(records, tail, None):
+            yield record
+    else:
+        with _record_buffer(records) as r:
+            record_count = sum(1 for record in r())
+            start_index = max(record_count - int(tail), 0)
+            for record in itertools.islice(r(), start_index, None):
+                yield record
 
 # Squeeze-related
 def gap_proportion(sequences, gap_chars='-'):
@@ -602,7 +662,7 @@ class CodonWarningTable(object):
             return '-'
         elif '-' in codon:
             if codon not in self.seen:
-                logging.warn("Unknown Codon: %s", codon)
+                logging.warning("Unknown Codon: %s", codon)
                 self.seen.add(codon)
             return self.missing_char
         else:
@@ -632,7 +692,9 @@ def translate(records, translate):
              'rna': CodonTable.ambiguous_rna_by_name["Standard"]}[source_type]
 
     # Handle ambiguities by replacing ambiguous codons with 'X'
-    table = copy.deepcopy(table)
+    # TODO: this copy operation causes infinite recursion with python3.6 -
+    # not sure why it was here to begin with.
+    # table = copy.deepcopy(table)
     table.forward_table = CodonWarningTable(table.forward_table)
 
     for record in records:
@@ -691,7 +753,7 @@ def sort_length(source_file, source_file_type, direction=1):
 
     # Adapted from the Biopython tutorial example.
 
-    #Get the lengths and ids, and sort on length
+    # Get the lengths and ids, and sort on length
     len_and_ids = sorted((len(rec), rec.id)
                          for rec in SeqIO.parse(source_file, source_file_type))
 
@@ -699,30 +761,49 @@ def sort_length(source_file, source_file_type, direction=1):
         ids = reversed([seq_id for (length, seq_id) in len_and_ids])
     else:
         ids = [seq_id for (length, seq_id) in len_and_ids]
-    del len_and_ids #free this memory
-    record_index = SeqIO.index(source_file.name, source_file_type)
-    records = (record_index[seq_id] for seq_id in ids)
+    del len_and_ids  # free this memory
 
-    return records
+    # SeqIO.index does not handle gzip instances
+    if isinstance(source_file, gzip.GzipFile):
+        tmpfile = tempfile.NamedTemporaryFile()
+        source_file.seek(0)
+        tmpfile.write(source_file.read())
+        tmpfile.seek(0)
+        source_file = tmpfile
+
+    record_index = SeqIO.index(source_file.name, source_file_type)
+
+    for seq_id in ids:
+        yield record_index[seq_id]
 
 
 def sort_name(source_file, source_file_type, direction=1):
     """
     Sort sequences by name. 1 is ascending (default) and 0 is descending.
     """
+
     direction_text = 'ascending' if direction == 1 else 'descending'
 
     logging.info("Indexing sequences by name: %s", direction_text)
 
     # Adapted from the Biopython tutorial example.
 
-    #Sort on id
+    # Sort on id
     ids = sorted((rec.id) for rec in SeqIO.parse(source_file,
                                                  source_file_type))
 
     if direction == 0:
         ids = reversed(ids)
-    record_index = SeqIO.index(source_file.name, source_file_type)
-    records = (record_index[id] for id in ids)
 
-    return records
+    # SeqIO.index does not handle gzip instances
+    if isinstance(source_file, gzip.GzipFile):
+        tmpfile = tempfile.NamedTemporaryFile()
+        source_file.seek(0)
+        tmpfile.write(source_file.read())
+        tmpfile.seek(0)
+        source_file = tmpfile
+
+    record_index = SeqIO.index(source_file.name, source_file_type)
+
+    for id in ids:
+        yield record_index[id]
